/*
 * File: macros.c
 * Date: 2011-11-17
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */

/***

In the original MicroEMACS, tokens of the form TK* were generally in the
top level of macro-handling.  These TK bits have been moved here.

***/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"macros.h"
#include	"bind.h"
#include	"buffer.h"
#include	"cmode.h"
#include	"cursor.h"
#include	"display.h"
#include	"envars.h"
#include	"exec.h"
#include	"killring.h"
#include	"line.h"
#include	"messages.h"
#include	"region.h"
#include	"search.h"
#include	"ufuncs.h"
#include	"terminal.h"
#include	"utils.h"

/*++

Find the TK* value of a token.
TK* are the elements of a macro (global_struct.h)

++*/
char *get_TKval(token)
char *token;

{
    int status;                 /* error return */
    BUFFER *bp;                 /* temp buffer pointer */
    int blen;                   /* length of buffer argument */
    static char buf[NSTRING];   /* string buffer for some returns */

    switch (get_TKtype(token)) {
    case TKNUL:
        return ("");

    case TKARG:                /* interactive argument */
        strcpy(token, fixnull(get_TKval(&token[1])));
        mlwrite("%s", token);
        status = mlgetstr(buf, NSTRING, ctoec(RETCHAR));
        if (status == ABORT)
            return (NULL);
        return (buf);

    case TKBUF:                /* buffer contents fetch */

        /* grab the right buffer */
        strcpy(token, fixnull(get_TKval(&token[1])));
        bp = exist_buffer(token, FALSE, 0);
        if (bp == NULL)
            return (NULL);

        /* if the buffer is displayed, get the window
           vars instead of the buffer vars */
        if (bp->b_nwnd > 0) {
            curbp->b_dotp = curwp->w_dotp;
            curbp->b_doto = curwp->w_doto;
        }

        /* if we are at the end, return <END> */
        if (bp->b_linep == bp->b_dotp)
            return ("<END>");

        /* grab the line as an argument */
        blen = lused(bp->b_dotp) - bp->b_doto;
        if (blen > NSTRING)
            blen = NSTRING;
        strncpy(buf, ltext(bp->b_dotp) + bp->b_doto, (size_t) blen);
        buf[blen] = 0;

        /* and step the buffer's line ptr ahead a line */
        bp->b_dotp = lforw(bp->b_dotp);
        bp->b_doto = 0;

        /* if displayed buffer, reset window ptr vars */
        if (bp->b_nwnd > 0) {
            curwp->w_dotp = curbp->b_dotp;
            curwp->w_doto = 0;
            curwp->w_flag |= WFMOVE;
        }

        /* and return the spoils */
        return (buf);

    case TKVAR:
        return (get_uvar(token + 1));
    case TKENV:
        return (eval_envar(token + 1));
    case TKFUN:
        return (eval_ufunc(token + 1));
    case TKDIR:
        return (NULL);
    case TKLBL:
        return (NULL);
    case TKLIT:
        return (token);
    case TKSTR:
        return (token + 1);
    case TKCMD:
        return (token);
    }
    return NULL;                /* should not get here */
}



/*++

Set tab size if given non-default argument (n <> 1).  Otherwise, insert a
tab into file.  If given argument, n, of zero, change to hard tabs.
If n > 1, simulate tab stop every n-characters using spaces. This has to be
done in this slightly funny way because the tab (in ASCII) has been turned
into "C-I" (in 10 bit code) already. Bound to "C-I".

++*/
int tab(f, n)
int f, n;
{
    f = 0;                      /* splint */

    if (n < 0)
        return (FALSE);

    if (n == 0 || n > 1) {      /* but not 1 */
        softtab = n;
        return (TRUE);
    }

    /* n == 1 */
    if (!softtab)
        return (insertnchars(1, '\t'));
    else
        return (insertnchars(softtab - (getccol(FALSE) % softtab), ' '));

}



/*++

Find the TK* type of a token.
TK* are the elements of a macro (global_struct.h)

++*/
int get_TKtype(token)
char *token;

{
    char c;                     /* first char in token */

    /* first char is all we need */
    c = *token;

    /* no blanks allowed */
    if (c == 0)
        return (TKNUL);

    /* a numeric literal */
    if (c >= '0' && c <= '9')
        return (TKLIT);

    switch (c) {
    case '"':
        return (TKSTR);
    case '!':
        return (TKDIR);
    case '@':
        return (TKARG);
    case '#':
        return (TKBUF);
    case '$':
        return (TKENV);
    case '%':
        return (TKVAR);
    case '&':
        return (TKFUN);
    case '*':
        return (TKLBL);

    default:
        return (TKCMD);
    }
}



/*++

Given char *var, looks for envar or user var of that name.
Helper for eval.c and main.c

++*/
void find_var(var, vd, size)
char *var;                      /* name of var to get */
VDESC *vd;                      /* structure to hold type and ptr */
int size;                       /* size of var array */

{
    int vnum;                   /* subscript in varable arrays */
    int vtype;                  /* type to return */

  fvar:vtype = -1;
    switch (var[0]) {

    case '$':                  /* check for legal environment var */
        if ((vnum = index_of_str(&var[1], get_envars_name, nenvars)) != -1)
            vtype = TKENV;
        break;

    case '%':                  /* check for existing legal user variable */
        for (vnum = 0; vnum < MAXUVARS; vnum++)
            if (strcmp(&var[1], uv[vnum].u_name) == 0) {
                vtype = TKVAR;
                break;
            }
        if (vnum < MAXUVARS)
            break;

        /* create a new one??? */
        for (vnum = 0; vnum < MAXUVARS; vnum++)
            if (uv[vnum].u_name[0] == 0) {
                vtype = TKVAR;
                strcpy(uv[vnum].u_name, &var[1]);
                uv[vnum].u_value = NULL;
                break;
            }
        break;

    case '&':                  /* indirect operator? */
        var[4] = 0;
        if (strcmp(&var[1], "ind") == 0) {
            /* grab token, and eval it */
            execstr = next_token(execstr, var, size);
            strcpy(var, fixnull(get_TKval(var)));
            goto fvar;
        }
    }

    /* return the results */
    vd->v_num = vnum;
    vd->v_type = vtype;
    return;
}



/*++

Set KEYTAB to the given name of the given type
Helper for set_var()

++*/
static int set_key(KEYTAB * key, char *name)
{
    int (*ktemp) ();            /* temp function pointer to assign */
    BUFFER *kmacro;             /* ptr to buffer of macro to bind to key */
    char bufn[LENPATH];         /* buffer to hold macro name */

    /* are we unbinding it? */
    if (*name == 0) {
        key->k_type = BINDNUL;
        return (TRUE);
    }

    /* bind to a built in function? */
    if ((ktemp = get_fn_name(name)) != NULL) {
        key->k_fp = ktemp;
        key->k_type = BINDFNC;
        return (TRUE);
    }

    /* is it a procedure/macro? */
    strcpy(bufn, "[");
    strcat(bufn, name);
    strcat(bufn, "]");
    if ((kmacro = exist_buffer(bufn, FALSE, 0)) != NULL) {
        key->k_buf = kmacro;
        key->k_type = BINDBUF;
        return (TRUE);
    }

    /* not anything we can bind to */
    mlwrite(TEXT16);            /*"[No such function]" */
    return (FALSE);
}



/*++

Given char *var, looks for envar or user var of that name and sets
it to char *val.
Helper for eval.c and main.c

++*/
int set_var(var, value)
VDESC *var;                     /* variable to set */
char *value;                    /* value to set to */

{
    int vnum;                   /* ordinal number of var refrenced */
    int vtype;                  /* type of variable to set */
    int status;                 /* status return */
    char *sp;                   /* scratch string pointer */

    /* simplify the vd structure (we are gonna look at it a lot) */
    vnum = var->v_num;
    vtype = var->v_type;

    /* and set the appropriate value */
    status = TRUE;
    switch (vtype) {
    case TKVAR:                /* set a user variable */
        if (uv[vnum].u_value != NULL)
            free(uv[vnum].u_value);
        sp = malloc(strlen(value) + 1);
        strcpy(sp, value);
        uv[vnum].u_value = sp;
        if (strcmp(value, errorm) == 0)
            status = FALSE;
        break;

    case TKENV:                /* set an environment variable */
        status = TRUE;          /* by default */

        switch (vnum) {
        case EVASAVE:
            asave = str2int(value);
            break;
        case EVBANGLINE:
            strcpy(bangline, value);
            break;
        case EVBUFHOOK:
            set_key(&bufhook, value);
            break;
        case EVCFNAME:
            strcpy(curbp->b_name, value);
            curwp->w_flag |= WFMODE;
            break;
        case EVCMDHK:
            set_key(&cmdhook, value);
            break;
        case EVCOMPILER:
            strcpy(compiler, value);
            break;
        case EVDEBUG:
            debug = str2bool(value);
            break;
        case EVEXBHOOK:
            set_key(&exbhook, value);
            break;
        case EVFILLCOL:
            fillcol = str2int(value);
            break;
        case EVFMTOPTS:
            strcpy(fmtopts, value);
            break;
        case EVHARDTAB:
            hardtab = str2int(value);
            update_window();
            break;
        case EVREADHK:
            set_key(&readhook, value);
            break;
        case EVREPLACE:
            strcpy((char *) rpat, value);
            rmcclear();
            break;
        case EVRVAL:
            break;
        case EVSEARCH:
            strcpy((char *) pat, value);
            setjtable();        /* Set up fast search arrays  */
            mcclear();
            break;
        case EVSOFTTAB:
            softtab = str2int(value);
            update_window();
            break;
        case EVSORTOPTS:
            strcpy(sortopts, value);
            break;
        case EVSTATUS:
            cmdstatus = str2bool(value);
            break;
        case EVTAGCMD:
            strcpy(tagcmd, value);
            break;
        case EVTIME:
            break;
        case EVWCHARS:
            setwlist(value);
            break;
        case EVWRAPHK:
            set_key(&wraphook, value);
            break;
        case EVWRITEHK:
            set_key(&writehook, value);
            break;
        }
        break;
    }
    return (status);
}




/*
 * Begin a keyboard macro.
 * Error if not at the top level in keyboard processing. Set up variables and
 * return.
 */

int begin_macro(f, n)
int f, n;

{

    f = n = 0;                  /* splint */

    if (kbdmode != STOP) {
        mlwrite(TEXT105);       /* "%%Macro already active" */
        return (FALSE);
    }
    mlwrite(TEXT106);           /*"[Start macro]" */
    kbdptr = &kbdm[0];
    kbdend = kbdptr;
    kbdmode = RECORD;
    return (TRUE);
}



/*
 * End keyboard macro. Check for the same limit conditions as the above
 * routine. Set up the variables and return to the caller.
 */

int end_macro(f, n)
int f, n;

{
    f = n = 0;                  /* splint */

    if (kbdmode == STOP) {
        mlwrite(TEXT107);       /*"%%Macro not active" */
        return (FALSE);
    }
    if (kbdmode == RECORD) {
        mlwrite(TEXT108);       /*"[End macro]" */
        kbdmode = STOP;
    }
    return (TRUE);
}



/*
 * Execute a macro.
 * The command argument is the number of times to loop. Quit as soon as a
 * command gets an error. Return TRUE if all ok, else FALSE.
 */

int execute_macro(f, n)
int f, n;

{

    f = 0;                      /* splint */

    if (kbdmode != STOP) {
        mlwrite(TEXT105);       /*"%%Macro already active" */
        return (FALSE);
    }
    if (n <= 0)
        return (TRUE);
    kbdrep = n;                 /* remember how many times to execute */
    kbdmode = PLAY;             /* start us in play mode */
    kbdptr = &kbdm[0];          /*    at the beginning */
    return (TRUE);
}



/*++

Twiddle the two characters on either side of dot. If dot is at the end of
the line twiddle the two characters before it. Return with an error if dot
is at the beginning of line; it seems to be a bit pointless to make this
work. This fixes up a very common typo with a single stroke. Normally bound
to "C-T". This always works within a line, so "WFEDIT" is good enough.

++*/
int twiddle(f, n)
int f, n;
{
    LINE *dotp;
    int doto;
    int cl;
    int cr;

    f = n = 0;                  /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    dotp = curwp->w_dotp;
    doto = curwp->w_doto;
    if (doto == lused(dotp) && --doto < 0)
        return (FALSE);
    cr = lgetc(dotp, doto);
    if (--doto < 0)
        return (FALSE);
    cl = lgetc(dotp, doto);
    lputc(dotp, doto + 0, cr);
    lputc(dotp, doto + 1, cl);
    change_line(WFEDIT);
    return (TRUE);
}



/*++

Quote the next character, and insert it into the buffer. All the
characters are taken literally, including the newline, which does not
then have its line splitting meaning. The character is always read, even
if it is inserted 0 times, for regularity. If a function key is pressed,
its symbolic name gets inserted!

++*/
int quote(f, n)
int f, n;
{
    int ec;                     /* current extended key fetched */
    int c;                      /* current ascii key fetched */
    int status;                 /* return value to hold from insertstring */
    char key_name[10];          /* name of a keystroke for quoting */

    f = 0;                      /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    ec = mlgetkey();

    /* fail this on a negative argument */
    if (n < 0)
        return (FALSE);

    /* nothing to insert . . . blow it off */
    if (n == 0)
        return (TRUE);

    /* if this is a function key, put its name in */
    if (ec & SPEC) {
        (void) cmd2str(ec, key_name);
        while (n--) {
            status = insertstring(key_name);
            if (status != TRUE)
                return (status);
        }
        return (TRUE);
    }

    /* otherwise, just insert the raw character */
    c = ectoc(ec);
    return (insertnchars(n, c));
}



/*++

Change tabs to spaces.

++*/
int detab(f, n)
int f, n;                       /* n = numeric repeat count */
{
    int inc;                    /* increment to next line [sgn(n)] */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */

    if (f == FALSE)
        n = linecountregion();

    /* loop thru detabbing n lines */
    inc = ((n > 0) ? 1 : -1);
    while (n) {
        curwp->w_doto = 0;      /* start at the beginning */

        /* detab the entire current line */
        while (curwp->w_doto < lused(curwp->w_dotp)) {
            /* if we have a tab */
            if (lgetc(curwp->w_dotp, curwp->w_doto) == '\t') {
                deletenchars(1L, FALSE);
                insertnchars(hardtab - (curwp->w_doto % hardtab), ' ');
            }
            next_character(FALSE, 1);
        }

        /* advance/or back to the next line */
        next_line(TRUE, inc);
        n -= inc;
    }
    curwp->w_doto = 0;          /* to the begining of the line */
    thisflag &= ~CFCPCN;        /* flag that this resets the goal column */
    change_line(WFEDIT);        /* yes, we have made at least an edit */
    return (TRUE);
}



/*++

Change spaces to tabs.

++*/
int entab(f, n)
int f, n;                       /* n = numeric repeat count */
{
    int inc;                    /* increment to next line [sgn(n)] */
    int fspace;                 /* pointer to first space if in a run */
    int ccol;                   /* current cursor column */
    char cchar;                 /* current character */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */

    if (f == FALSE)
        n = linecountregion();

    /* loop thru entabbing n lines */
    inc = ((n > 0) ? 1 : -1);
    while (n) {
        /* entab the entire current line */

        ccol = curwp->w_doto = 0;       /* start at the beginning */
        fspace = -1;

        while (curwp->w_doto < lused(curwp->w_dotp)) {
            /* see if it is time to compress */
            if ((fspace >= 0) && (nextab(fspace) <= ccol)) {
                if (ccol - fspace < 2)
                    fspace = -1;
                else {
                    previous_character(TRUE, ccol - fspace);
                    deletenchars((long) (ccol - fspace), FALSE);
                    insertnchars(1, '\t');
                    fspace = -1;
                }
            }

            /* get the current character */
            cchar = lgetc(curwp->w_dotp, curwp->w_doto);

            switch (cchar) {
            case '\t':         /* a tab...count em up (no break here)  */
                deletenchars(1L, FALSE);
                insertnchars(hardtab - (ccol % hardtab), ' ');

            case ' ':          /* a space...compress? */
                if (fspace == -1)
                    fspace = ccol;
                break;

            default:           /* any other char...just count */
                fspace = -1;
                break;
            }
            ccol++;
            next_character(FALSE, 1);
        }

        /* advance/or back to the next line */
        next_line(TRUE, inc);
        n -= inc;
        curwp->w_doto = 0;      /* start at the beginning */
    }
    curwp->w_doto = 0;          /* to the begining of the line */
    thisflag &= ~CFCPCN;        /* flag that this resets the goal column */
    change_line(WFEDIT);        /* yes, we have made at least an edit */
    return (TRUE);
}



/*++

Trim trailing whitespace from the point to eol with no arguments, it
trims the current region.

++*/
int trim(f, n)
int f, n;                       /* n = numeric repeat count */
{
    LINE *lp;                   /* current line pointer */
    int offset;                 /* original line offset position */
    int length;                 /* current length */
    int inc;                    /* increment to next line [sgn(n)] */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */

    if (f == FALSE)
        n = linecountregion();

    /* loop thru trimming n lines */
    inc = ((n > 0) ? 1 : -1);
    while (n) {
        lp = curwp->w_dotp;     /* find current line text */
        offset = curwp->w_doto; /* save original offset */
        length = lused(lp);     /* find current length */

        /* trim the current line */
        while (length > offset) {
            if (lgetc(lp, length - 1) != ' '
                && lgetc(lp, length - 1) != '\t')
                break;
            length--;
        }
        lp->l_used = (size_t) length;

        /* advance/or back to the next line */
        next_line(TRUE, inc);
        n -= inc;
    }
    change_line(WFEDIT);
    thisflag &= ~CFCPCN;        /* flag that this resets the goal column */
    return (TRUE);
}



/*++

Open up some blank space. The basic plan is to insert a bunch of newlines,
and then back up over them. 

++*/
int open_line(f, n)
int f, n;
{
    int i;
    int s;

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if (n < 0)
        return (FALSE);
    if (n == 0)
        return (TRUE);
    i = n;                      /* Insert newlines.     */
    do {
        s = insertnewline();
    }
    while (s == TRUE && --i);
    if (s == TRUE)              /* Then back up overtop */
        s = previous_character(f, n);   /* of them all.         */
    return (s);
}




/*++

Insert a newline. Bound to "C-M" which is [Enter]. If we are in CMODE,
do automatic indentation as specified.

++*/
int newline(f, n)
int f, n;
{
    int s;

    f = 0;                      /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if (n < 0)
        return (FALSE);

    /* if we are in C mode and this is a default <NL> */
    if (n == 1 && (curbp->b_mode & MDCMOD)
        && curwp->w_dotp != curbp->b_linep)
        return (cinsert());

    /*
     * If a newline was typed, fill column is defined, the argument is non-
     * negative, wrap mode is enabled, and we are now past fill column,
     * and we are not read-only, perform word wrap.
     */
    if ((curwp->w_bufp->b_mode & MDWRAP) && fillcol > 0 &&
        getccol(FALSE) > fillcol
        && (curwp->w_bufp->b_mode & MDVIEW) == FALSE)
        execute_key(&wraphook, FALSE, 1);

    /* insert some lines */
    while (n--) {
        if ((s = insertnewline()) != TRUE)
            return (s);
    }
    return (TRUE);
}


/*++

Delete forward. This is real easy, because the basic delete routine does
all of the work. Watches for negative arguments, and does the right thing.
If any argument is present, it kills rather than deletes, to prevent loss
of text if typed with a big argument. Normally bound to "C-D".

++*/
int delete_forward(f, n)
int f, n;
{
    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if (n < 0)
        return (delete_backward(f, -n));
    if (f != FALSE) {           /* Really a kill. */
        if ((lastflag & CFKILL) == 0)
            next_kill();
        thisflag |= CFKILL;
    }
    return (deletenchars((long) n, f));
}

/*++

Delete backward. This is quite easy too, because it's all done with other
functions. Just move the cursor back, and delete forwards. Like delete
forward, this actually does a kill if presented with an argument. Bound to
both "RUBOUT" and "C-H".

++*/
int delete_backward(f, n)
int f, n;
{
    int s;

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if (n < 0)
        return (delete_forward(f, -n));
    if (f != FALSE) {           /* Really a kill.       */
        if ((lastflag & CFKILL) == 0)
            next_kill();
        thisflag |= CFKILL;
    }
    if ((s = previous_character(f, n)) == TRUE)
        s = deletenchars((long) n, f);
    return (s);
}



/*++

Ask for and insert a string into the current buffer at the current point

++*/
int insert_string(f, n)
int f, n;
{
    int status;                 /* status return code */
    char tstring[NPAT + 1];     /* string to add */

    /* ask for string to insert */
    status = next_arg(TEXT68, tstring, NPAT, sterm);    /*"String to insert: " */
    if (status != TRUE)
        return (status);

    if (f == FALSE)
        n = 1;

    if (n < 0)
        n = -n;

    /* insert it */
    while (n-- && (status = insertstring(tstring)));
    return (status);
}
